"""This is part of the MSS Python's module.
Source: https://github.com/BoboTiG/python-mss.
"""

import os
import platform
from collections.abc import Callable

import pytest

import mss

OS = platform.system().lower()
PID = os.getpid()


def get_opened_socket() -> int:
    """GNU/Linux: a way to get the opened sockets count.
    It will be used to check X server connections are well closed.
    """
    import subprocess

    cmd = f"lsof -U | grep {PID}"
    output = subprocess.check_output(cmd, shell=True)
    return len(output.splitlines())


def get_handles() -> int:
    """Windows: a way to get the GDI handles count.
    It will be used to check the handles count is not growing, showing resource leaks.
    """
    import ctypes

    PROCESS_QUERY_INFORMATION = 0x400  # noqa:N806
    GR_GDIOBJECTS = 0  # noqa:N806
    h = ctypes.windll.kernel32.OpenProcess(PROCESS_QUERY_INFORMATION, 0, PID)
    return ctypes.windll.user32.GetGuiResources(h, GR_GDIOBJECTS)


@pytest.fixture
def monitor_func() -> Callable[[], int]:
    """OS specific function to check resources in use."""
    return get_opened_socket if OS == "linux" else get_handles


def bound_instance_without_cm() -> None:
    # Will always leak
    sct = mss.mss()
    sct.shot()


def bound_instance_without_cm_but_use_close() -> None:
    sct = mss.mss()
    sct.shot()
    sct.close()
    # Calling .close() twice should be possible
    sct.close()


def unbound_instance_without_cm() -> None:
    # Will always leak
    mss.mss().shot()


def with_context_manager() -> None:
    with mss.mss() as sct:
        sct.shot()


def regression_issue_128() -> None:
    """Regression test for issue #128: areas overlap."""
    with mss.mss() as sct:
        area1 = {"top": 50, "left": 7, "width": 400, "height": 320, "mon": 1}
        sct.grab(area1)
        area2 = {"top": 200, "left": 200, "width": 320, "height": 320, "mon": 1}
        sct.grab(area2)


def regression_issue_135() -> None:
    """Regression test for issue #135: multiple areas."""
    with mss.mss() as sct:
        bounding_box_notes = {"top": 0, "left": 0, "width": 100, "height": 100}
        sct.grab(bounding_box_notes)
        bounding_box_test = {"top": 220, "left": 220, "width": 100, "height": 100}
        sct.grab(bounding_box_test)
        bounding_box_score = {"top": 110, "left": 110, "width": 100, "height": 100}
        sct.grab(bounding_box_score)


def regression_issue_210() -> None:
    """Regression test for issue #210: multiple X servers."""
    pyvirtualdisplay = pytest.importorskip("pyvirtualdisplay")

    with pyvirtualdisplay.Display(size=(1920, 1080), color_depth=24), mss.mss():
        pass

    with pyvirtualdisplay.Display(size=(1920, 1080), color_depth=24), mss.mss():
        pass


@pytest.mark.skipif(OS == "darwin", reason="No possible leak on macOS.")
@pytest.mark.parametrize(
    "func",
    [
        # bound_instance_without_cm,
        bound_instance_without_cm_but_use_close,
        # unbound_instance_without_cm,
        with_context_manager,
        regression_issue_128,
        regression_issue_135,
        regression_issue_210,
    ],
)
def test_resource_leaks(func: Callable[[], None], monitor_func: Callable[[], int]) -> None:
    """Check for resource leaks with different use cases."""
    # Warm-up
    func()

    original_resources = monitor_func()
    allocated_resources = 0

    for _ in range(5):
        func()
        new_resources = monitor_func()
        allocated_resources = max(allocated_resources, new_resources)

    assert allocated_resources <= original_resources



"""Smart Detection Example

This example demonstrates how to use PyShotter's smart detection features.
"""

from pyshotter import pyshotter, SmartDetectionFeature, AnnotationFeature

def detect_and_annotate():
    """Detect code regions and windows, then annotate them."""
    
    with pyshotter() as sct:
        # Take a screenshot
        screenshot = sct.grab(sct.monitors[0])
        
        # Initialize smart detection
        detector = SmartDetectionFeature()
        
        # Detect code regions
        print("🔍 Detecting code regions...")
        code_regions = detector.detect_code_regions(screenshot)
        print(f"Found {len(code_regions)} code regions")
        
        # Detect windows
        print("🖥️ Detecting windows...")
        windows = detector.detect_windows(screenshot)
        print(f"Found {len(windows)} windows")
        
        # Annotate detected regions
        annotator = AnnotationFeature()
        annotated = screenshot
        
        # Highlight code regions with rectangles
        for i, region in enumerate(code_regions[:5]):  # Limit to first 5
            x, y, w, h = region['bbox']
            annotated = annotator.add_rectangle(
                annotated,
                (x, y),
                (x + w, y + h),
                color=(0, 255, 0),  # Green for code
                thickness=2
            )
            # Add label
            annotated = annotator.add_text(
                annotated,
                f"Code {i+1}",
                (x, y - 20),
                font_size=16,
                color=(0, 255, 0)
            )
        
        # Highlight windows with circles
        for i, window in enumerate(windows[:3]):  # Limit to first 3
            x, y, w, h = window['bbox']
            center_x = x + w // 2
            center_y = y + h // 2
            radius = min(w, h) // 4
            
            annotated = annotator.add_circle(
                annotated,
                (center_x, center_y),
                radius,
                color=(255, 0, 0),  # Red for windows
                thickness=2
            )
            # Add label
            annotated = annotator.add_text(
                annotated,
                f"Window {i+1}",
                (x, y - 20),
                font_size=16,
                color=(255, 0, 0)
            )
        
        # Save the annotated screenshot
        annotated.save("smart_detection_result.png")
        print("✅ Smart detection result saved as 'smart_detection_result.png'")
        
        return code_regions, windows

def analyze_screenshot():
    """Analyze screenshot content and provide insights."""
    
    with pyshotter() as sct:
        screenshot = sct.grab(sct.monitors[0])
        detector = SmartDetectionFeature()
        
        # Get detection results
        code_regions = detector.detect_code_regions(screenshot)
        windows = detector.detect_windows(screenshot)
        
        # Analyze and report
        print("📊 Screenshot Analysis")
        print("=" * 30)
        print(f"📏 Screenshot size: {screenshot.size}")
        print(f"🔍 Code regions detected: {len(code_regions)}")
        print(f"🖥️ Windows detected: {len(windows)}")
        
        if code_regions:
            total_code_area = sum(region['area'] for region in code_regions)
            print(f"📝 Total code area: {total_code_area} pixels")
        
        if windows:
            largest_window = max(windows, key=lambda w: w['area'])
            print(f"🖼️ Largest window: {largest_window['bbox']}")
        
        # Provide recommendations
        print("\n💡 Recommendations:")
        if len(code_regions) > 3:
            print("  - Consider annotating important code sections")
        if len(windows) > 2:
            print("  - Multiple windows detected - consider panorama mode")
        if not code_regions and not windows:
            print("  - No specific content detected - try different detection settings")

if __name__ == "__main__":
    print("🧠 PyShotter Smart Detection Demo")
    print("=" * 50)
    
    # Run detection and annotation
    code_regions, windows = detect_and_annotate()
    
    print("\n" + "=" * 50)
    
    # Run analysis
    analyze_screenshot() 


"""Sensitive Data Redaction Example

This example demonstrates how to automatically redact sensitive information from screenshots.
"""

from pyshotter import pyshotter, RedactionFeature

def redact_sensitive_data():
    """Redact sensitive data from a screenshot."""
    
    # Take a screenshot
    with pyshotter() as sct:
        screenshot = sct.grab(sct.monitors[0])
        
        # Initialize redaction feature
        redaction = RedactionFeature()
        
        # Redact sensitive data
        clean_screenshot = redaction.redact_sensitive_data(screenshot)
        
        # Save the redacted screenshot
        clean_screenshot.save("redacted_screenshot.png")
        print("Sensitive data has been redacted and saved to 'redacted_screenshot.png'")

if __name__ == "__main__":
    redact_sensitive_data() 


"""Advanced features for PyShotter.

This module provides advanced screenshot features including:
- OCR text extraction
- Sensitive data redaction
- Multi-monitor panorama
- Change detection
- Hotkey management
- Screenshot history and search
- Annotation tools
- Sharing capabilities
"""

import os
import re
import time
import json
import base64
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

try:
    import cv2
    import numpy as np
    OPENCV_AVAILABLE = True
except ImportError:
    OPENCV_AVAILABLE = False

try:
    import pytesseract
    TESSERACT_AVAILABLE = True
except ImportError:
    TESSERACT_AVAILABLE = False

from .screenshot import ScreenShot


class AnnotationFeature:
    """Smart annotation tools for screenshots."""
    
    def __init__(self):
        if not OPENCV_AVAILABLE:
            raise ImportError("opencv-python is required for annotation features. Install with: pip install opencv-python")
    
    def add_text(self, screenshot: ScreenShot, text: str, position: Tuple[int, int], 
                 font_size: int = 20, color: Tuple[int, int, int] = (255, 0, 0)) -> ScreenShot:
        """Add text annotation to a screenshot.
        
        Args:
            screenshot: The screenshot to annotate
            text: Text to add
            position: (x, y) position for text
            font_size: Font size
            color: RGB color tuple
            
        Returns:
            Annotated screenshot
        """
        img_array = np.array(screenshot.rgb).reshape(screenshot.height, screenshot.width, 3)
        
        # Add text using OpenCV
        cv2.putText(img_array, text, position, cv2.FONT_HERSHEY_SIMPLEX, 
                   font_size/30, color, 2)
        
        annotated_rgb = img_array.tobytes()
        return ScreenShot(
            rgb=annotated_rgb,
            size=screenshot.size,
            pos=screenshot.pos
        )
    
    def add_rectangle(self, screenshot: ScreenShot, top_left: Tuple[int, int], 
                     bottom_right: Tuple[int, int], color: Tuple[int, int, int] = (255, 0, 0),
                     thickness: int = 2) -> ScreenShot:
        """Add rectangle annotation to a screenshot.
        
        Args:
            screenshot: The screenshot to annotate
            top_left: Top-left corner (x, y)
            bottom_right: Bottom-right corner (x, y)
            color: RGB color tuple
            thickness: Line thickness
            
        Returns:
            Annotated screenshot
        """
        img_array = np.array(screenshot.rgb).reshape(screenshot.height, screenshot.width, 3)
        
        cv2.rectangle(img_array, top_left, bottom_right, color, thickness)
        
        annotated_rgb = img_array.tobytes()
        return ScreenShot(
            rgb=annotated_rgb,
            size=screenshot.size,
            pos=screenshot.pos
        )
    
    def add_arrow(self, screenshot: ScreenShot, start: Tuple[int, int], end: Tuple[int, int],
                  color: Tuple[int, int, int] = (255, 0, 0), thickness: int = 2) -> ScreenShot:
        """Add arrow annotation to a screenshot.
        
        Args:
            screenshot: The screenshot to annotate
            start: Arrow start point (x, y)
            end: Arrow end point (x, y)
            color: RGB color tuple
            thickness: Line thickness
            
        Returns:
            Annotated screenshot
        """
        img_array = np.array(screenshot.rgb).reshape(screenshot.height, screenshot.width, 3)
        
        cv2.arrowedLine(img_array, start, end, color, thickness)
        
        annotated_rgb = img_array.tobytes()
        return ScreenShot(
            rgb=annotated_rgb,
            size=screenshot.size,
            pos=screenshot.pos
        )
    
    def add_circle(self, screenshot: ScreenShot, center: Tuple[int, int], radius: int,
                   color: Tuple[int, int, int] = (255, 0, 0), thickness: int = 2) -> ScreenShot:
        """Add circle annotation to a screenshot.
        
        Args:
            screenshot: The screenshot to annotate
            center: Circle center (x, y)
            radius: Circle radius
            color: RGB color tuple
            thickness: Line thickness (-1 for filled)
            
        Returns:
            Annotated screenshot
        """
        img_array = np.array(screenshot.rgb).reshape(screenshot.height, screenshot.width, 3)
        
        cv2.circle(img_array, center, radius, color, thickness)
        
        annotated_rgb = img_array.tobytes()
        return ScreenShot(
            rgb=annotated_rgb,
            size=screenshot.size,
            pos=screenshot.pos
        )
    
    def add_highlight(self, screenshot: ScreenShot, region: Tuple[int, int, int, int],
                      color: Tuple[int, int, int] = (255, 255, 0), alpha: float = 0.3) -> ScreenShot:
        """Add highlight overlay to a region.
        
        Args:
            screenshot: The screenshot to annotate
            region: (x, y, width, height) region to highlight
            color: RGB color tuple
            alpha: Transparency (0.0 to 1.0)
            
        Returns:
            Annotated screenshot
        """
        img_array = np.array(screenshot.rgb).reshape(screenshot.height, screenshot.width, 3)
        x, y, w, h = region
        
        # Create overlay
        overlay = img_array.copy()
        cv2.rectangle(overlay, (x, y), (x + w, y + h), color, -1)
        
        # Blend with original
        cv2.addWeighted(overlay, alpha, img_array, 1 - alpha, 0, img_array)
        
        annotated_rgb = img_array.tobytes()
        return ScreenShot(
            rgb=annotated_rgb,
            size=screenshot.size,
            pos=screenshot.pos
        )


class SharingFeature:
    """Smart sharing capabilities for screenshots."""
    
    def __init__(self):
        self.clipboard_available = self._check_clipboard()
    
    def _check_clipboard(self) -> bool:
        """Check if clipboard functionality is available."""
        try:
            import pyperclip
            return True
        except ImportError:
            return False
    
    def copy_to_clipboard(self, screenshot: ScreenShot) -> bool:
        """Copy screenshot to clipboard.
        
        Args:
            screenshot: Screenshot to copy
            
        Returns:
            True if successful, False otherwise
        """
        if not self.clipboard_available:
            return False
        
        try:
            import pyperclip
            from PIL import Image
            import io
            import base64
            
            # Convert to PIL Image
            img = Image.frombytes("RGB", screenshot.size, screenshot.rgb)
            
            # Save to bytes
            img_bytes = io.BytesIO()
            img.save(img_bytes, format='PNG')
            img_bytes.seek(0)
            
            # Copy to clipboard
            pyperclip.copy(img_bytes.getvalue())
            return True
        except Exception:
            return False
    
    def generate_shareable_link(self, screenshot: ScreenShot, service: str = "imgur") -> Optional[str]:
        """Generate a shareable link for the screenshot.
        
        Args:
            screenshot: Screenshot to share
            service: Sharing service ('imgur', 'pastebin', etc.)
            
        Returns:
            Shareable URL or None if failed
        """
        try:
            from PIL import Image
            import io
            import base64
            
            # Convert to PIL Image
            img = Image.frombytes("RGB", screenshot.size, screenshot.rgb)
            
            # Save to bytes
            img_bytes = io.BytesIO()
            img.save(img_bytes, format='PNG')
            img_bytes.seek(0)
            
            # Encode to base64
            img_base64 = base64.b64encode(img_bytes.getvalue()).decode()
            
            if service == "imgur":
                # For demo purposes, return a data URL
                return f"data:image/png;base64,{img_base64}"
            else:
                return f"data:image/png;base64,{img_base64}"
                
        except Exception:
            return None
    
    def save_with_metadata(self, screenshot: ScreenShot, filename: str, 
                          metadata: Dict[str, Any] = None) -> bool:
        """Save screenshot with embedded metadata.
        
        Args:
            screenshot: Screenshot to save
            filename: Output filename
            metadata: Metadata to embed
            
        Returns:
            True if successful, False otherwise
        """
        try:
            from PIL import Image, PngInfo
            
            # Convert to PIL Image
            img = Image.frombytes("RGB", screenshot.size, screenshot.rgb)
            
            # Add metadata
            if metadata:
                pnginfo = PngInfo()
                for key, value in metadata.items():
                    pnginfo.add_text(key, str(value))
                img.save(filename, "PNG", pnginfo=pnginfo)
            else:
                img.save(filename, "PNG")
            
            return True
        except Exception:
            return False


class SmartDetectionFeature:
    """Smart detection features for screenshots."""
    
    def __init__(self):
        if not OPENCV_AVAILABLE:
            raise ImportError("opencv-python is required for smart detection. Install with: pip install opencv-python")
    
    def detect_code_regions(self, screenshot: ScreenShot) -> List[Dict[str, Any]]:
        """Detect code-like regions in screenshots.
        
        Args:
            screenshot: Screenshot to analyze
            
        Returns:
            List of detected code regions with bounding boxes
        """
        img_array = np.array(screenshot.rgb).reshape(screenshot.height, screenshot.width, 3)
        
        # Convert to grayscale
        gray = cv2.cvtColor(img_array, cv2.COLOR_RGB2GRAY)
        
        # Detect text-like regions (simple approach)
        # In a real implementation, you'd use more sophisticated OCR/text detection
        _, binary = cv2.threshold(gray, 0, 255, cv2.THRESH_BINARY_INV + cv2.THRESH_OTSU)
        
        # Find contours
        contours, _ = cv2.findContours(binary, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
        
        code_regions = []
        for contour in contours:
            x, y, w, h = cv2.boundingRect(contour)
            if w > 50 and h > 20:  # Filter small regions
                code_regions.append({
                    'bbox': (x, y, w, h),
                    'area': w * h,
                    'confidence': 0.8  # Placeholder confidence
                })
        
        return code_regions
    
    def detect_windows(self, screenshot: ScreenShot) -> List[Dict[str, Any]]:
        """Detect application windows in screenshots.
        
        Args:
            screenshot: Screenshot to analyze
            
        Returns:
            List of detected windows with bounding boxes
        """
        img_array = np.array(screenshot.rgb).reshape(screenshot.height, screenshot.width, 3)
        
        # Convert to grayscale
        gray = cv2.cvtColor(img_array, cv2.COLOR_RGB2GRAY)
        
        # Detect edges
        edges = cv2.Canny(gray, 50, 150)
        
        # Find contours
        contours, _ = cv2.findContours(edges, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
        
        windows = []
        for contour in contours:
            x, y, w, h = cv2.boundingRect(contour)
            if w > 100 and h > 50:  # Filter small regions
                windows.append({
                    'bbox': (x, y, w, h),
                    'area': w * h
                })
        
        return windows


class OCRFeature:
    """OCR (Optical Character Recognition) feature for extracting text from screenshots."""
    
    def __init__(self):
        if not TESSERACT_AVAILABLE:
            raise ImportError("pytesseract is required for OCR features. Install with: pip install pytesseract")
    
    def extract_text(self, screenshot: ScreenShot, lang: str = 'eng') -> str:
        """Extract text from a screenshot using OCR.
        
        Args:
            screenshot: The screenshot to extract text from
            lang: Language code for OCR (default: 'eng')
            
        Returns:
            Extracted text as string
        """
        # Convert screenshot to PIL Image for OCR
        from PIL import Image
        img = Image.frombytes("RGB", screenshot.size, screenshot.rgb)
        
        # Extract text using Tesseract
        text = pytesseract.image_to_string(img, lang=lang)
        return text.strip()
    
    def extract_text_boxes(self, screenshot: ScreenShot, lang: str = 'eng') -> List[Dict[str, Any]]:
        """Extract text with bounding boxes from a screenshot.
        
        Args:
            screenshot: The screenshot to extract text from
            lang: Language code for OCR (default: 'eng')
            
        Returns:
            List of dictionaries with text and bounding box information
        """
        from PIL import Image
        img = Image.frombytes("RGB", screenshot.size, screenshot.rgb)
        
        # Get detailed OCR data
        data = pytesseract.image_to_data(img, lang=lang, output_type=pytesseract.Output.DICT)
        
        results = []
        for i in range(len(data['text'])):
            if data['conf'][i] > 0:  # Filter out low confidence results
                results.append({
                    'text': data['text'][i],
                    'confidence': data['conf'][i],
                    'bbox': (data['left'][i], data['top'][i], 
                            data['left'][i] + data['width'][i], 
                            data['top'][i] + data['height'][i])
                })
        
        return results


class RedactionFeature:
    """Sensitive data redaction feature for screenshots."""
    
    def __init__(self):
        if not OPENCV_AVAILABLE:
            raise ImportError("opencv-python is required for redaction features. Install with: pip install opencv-python")
    
    def redact_sensitive_data(self, screenshot: ScreenShot) -> ScreenShot:
        """Redact sensitive data from a screenshot.
        
        Args:
            screenshot: The screenshot to redact
            
        Returns:
            New screenshot with sensitive data redacted
        """
        # Convert to numpy array
        img_array = np.array(screenshot.rgb).reshape(screenshot.height, screenshot.width, 3)
        
        # Extract text to find sensitive data
        ocr = OCRFeature()
        text_boxes = ocr.extract_text_boxes(screenshot)
        
        # Patterns for sensitive data
        patterns = {
            'email': r'\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}\b',
            'phone': r'\b\d{3}[-.]?\d{3}[-.]?\d{4}\b',
            'credit_card': r'\b\d{4}[- ]?\d{4}[- ]?\d{4}[- ]?\d{4}\b',
            'ssn': r'\b\d{3}-\d{2}-\d{4}\b'
        }
        
        # Redact sensitive data
        for box in text_boxes:
            text = box['text']
            for pattern_name, pattern in patterns.items():
                if re.search(pattern, text):
                    # Blur the bounding box area
                    x1, y1, x2, y2 = box['bbox']
                    img_array[y1:y2, x1:x2] = cv2.GaussianBlur(
                        img_array[y1:y2, x1:x2], (15, 15), 0
                    )
        
        # Convert back to screenshot format
        redacted_rgb = img_array.tobytes()
        return ScreenShot(
            rgb=redacted_rgb,
            size=screenshot.size,
            pos=screenshot.pos
        )


class PanoramaFeature:
    """Multi-monitor panorama feature for stitching screenshots."""
    
    def create_panorama(self, screenshots: List[ScreenShot]) -> ScreenShot:
        """Create a panoramic image from multiple monitor screenshots.
        
        Args:
            screenshots: List of screenshots from different monitors
            
        Returns:
            Panoramic screenshot
        """
        if not screenshots:
            raise ValueError("At least one screenshot is required")
        
        # Calculate total width and max height
        total_width = sum(s.width for s in screenshots)
        max_height = max(s.height for s in screenshots)
        
        # Create panoramic image
        panorama_array = np.zeros((max_height, total_width, 3), dtype=np.uint8)
        
        x_offset = 0
        for screenshot in screenshots:
            img_array = np.array(screenshot.rgb).reshape(screenshot.height, screenshot.width, 3)
            panorama_array[:screenshot.height, x_offset:x_offset + screenshot.width] = img_array
            x_offset += screenshot.width
        
        # Convert back to screenshot format
        panorama_rgb = panorama_array.tobytes()
        return ScreenShot(
            rgb=panorama_rgb,
            size=(total_width, max_height),
            pos=(0, 0)
        )


class ChangeDetectionFeature:
    """Change detection feature for comparing screenshots."""
    
    def __init__(self):
        if not OPENCV_AVAILABLE:
            raise ImportError("opencv-python is required for change detection. Install with: pip install opencv-python")
    
    def detect_changes(self, current: ScreenShot, previous: ScreenShot, 
                      threshold: float = 0.1) -> ScreenShot:
        """Detect changes between two screenshots.
        
        Args:
            current: Current screenshot
            previous: Previous screenshot
            threshold: Sensitivity threshold for change detection
            
        Returns:
            Screenshot highlighting changes
        """
        # Convert to numpy arrays
        current_array = np.array(current.rgb).reshape(current.height, current.width, 3)
        previous_array = np.array(previous.rgb).reshape(previous.height, previous.width, 3)
        
        # Convert to grayscale for comparison
        current_gray = cv2.cvtColor(current_array, cv2.COLOR_RGB2GRAY)
        previous_gray = cv2.cvtColor(previous_array, cv2.COLOR_RGB2GRAY)
        
        # Calculate difference
        diff = cv2.absdiff(current_gray, previous_gray)
        
        # Apply threshold
        _, thresh = cv2.threshold(diff, int(255 * threshold), 255, cv2.THRESH_BINARY)
        
        # Find contours of changes
        contours, _ = cv2.findContours(thresh, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
        
        # Draw red rectangles around changes
        result_array = current_array.copy()
        for contour in contours:
            if cv2.contourArea(contour) > 100:  # Filter small changes
                x, y, w, h = cv2.boundingRect(contour)
                cv2.rectangle(result_array, (x, y), (x + w, y + h), (255, 0, 0), 2)
        
        # Convert back to screenshot format
        result_rgb = result_array.tobytes()
        return ScreenShot(
            rgb=result_rgb,
            size=current.size,
            pos=current.pos
        )


class HotkeyManager:
    """Global hotkey manager for screenshot triggers."""
    
    def __init__(self):
        self.hotkeys = {}
        self._running = False
    
    def register_hotkey(self, key_combination: str, callback: callable) -> None:
        """Register a global hotkey.
        
        Args:
            key_combination: Hotkey combination (e.g., 'ctrl+shift+s')
            callback: Function to call when hotkey is pressed
        """
        self.hotkeys[key_combination] = callback
    
    def unregister_hotkey(self, key_combination: str) -> None:
        """Unregister a global hotkey.
        
        Args:
            key_combination: Hotkey combination to unregister
        """
        if key_combination in self.hotkeys:
            del self.hotkeys[key_combination]
    
    def start_listening(self) -> None:
        """Start listening for hotkeys."""
        # Implementation would depend on platform-specific libraries
        # For now, this is a placeholder
        self._running = True
    
    def stop_listening(self) -> None:
        """Stop listening for hotkeys."""
        self._running = False


class ScreenshotHistory:
    """Screenshot history and search functionality."""
    
    def __init__(self, history_dir: str = "~/.pyshotter/history"):
        self.history_dir = Path(history_dir).expanduser()
        self.history_dir.mkdir(parents=True, exist_ok=True)
        self.history_file = self.history_dir / "history.json"
        self._load_history()
    
    def _load_history(self) -> None:
        """Load screenshot history from file."""
        import json
        if self.history_file.exists():
            with open(self.history_file, 'r') as f:
                self.history = json.load(f)
        else:
            self.history = []
    
    def _save_history(self) -> None:
        """Save screenshot history to file."""
        import json
        with open(self.history_file, 'w') as f:
            json.dump(self.history, f, indent=2)
    
    def add_screenshot(self, screenshot: ScreenShot, metadata: Dict[str, Any] = None) -> str:
        """Add a screenshot to history.
        
        Args:
            screenshot: The screenshot to add
            metadata: Additional metadata (window title, tags, etc.)
            
        Returns:
            Screenshot ID
        """
        import json
        import hashlib
        
        # Generate unique ID
        screenshot_id = hashlib.md5(f"{time.time()}".encode()).hexdigest()[:8]
        
        # Save screenshot
        screenshot_path = self.history_dir / f"{screenshot_id}.png"
        screenshot.save(str(screenshot_path))
        
        # Extract text if OCR is available
        ocr_text = ""
        if TESSERACT_AVAILABLE:
            try:
                ocr = OCRFeature()
                ocr_text = ocr.extract_text(screenshot)
            except:
                pass
        
        # Create history entry
        entry = {
            'id': screenshot_id,
            'timestamp': datetime.now().isoformat(),
            'path': str(screenshot_path),
            'size': screenshot.size,
            'ocr_text': ocr_text,
            'metadata': metadata or {}
        }
        
        self.history.append(entry)
        self._save_history()
        
        return screenshot_id
    
    def search_history(self, query: str) -> List[Dict[str, Any]]:
        """Search screenshot history.
        
        Args:
            query: Search query (searches in OCR text, metadata, and tags)
            
        Returns:
            List of matching screenshot entries
        """
        results = []
        query_lower = query.lower()
        
        for entry in self.history:
            # Search in OCR text
            if query_lower in entry.get('ocr_text', '').lower():
                results.append(entry)
                continue
            
            # Search in metadata
            metadata_str = json.dumps(entry.get('metadata', {})).lower()
            if query_lower in metadata_str:
                results.append(entry)
                continue
            
            # Search in tags
            tags = entry.get('metadata', {}).get('tags', [])
            if any(query_lower in tag.lower() for tag in tags):
                results.append(entry)
        
        return results
    
    def get_screenshot(self, screenshot_id: str) -> Optional[ScreenShot]:
        """Get a screenshot from history by ID.
        
        Args:
            screenshot_id: ID of the screenshot to retrieve
            
        Returns:
            Screenshot object or None if not found
        """
        for entry in self.history:
            if entry['id'] == screenshot_id:
                screenshot_path = Path(entry['path'])
                if screenshot_path.exists():
                    # Load screenshot from file
                    # This would need to be implemented based on your screenshot loading method
                    return None  # Placeholder
        return None 


"""This is part of the MSS Python's module.
Source: https://github.com/BoboTiG/python-mss.
"""

import os
from typing import Any

from mss import mss
from mss.models import Monitor


class SimpleScreenShot:
    def __init__(self, data: bytearray, monitor: Monitor, **_: Any) -> None:
        self.raw = bytes(data)
        self.monitor = monitor


def test_custom_cls_image() -> None:
    with mss(display=os.getenv("DISPLAY")) as sct:
        sct.cls_image = SimpleScreenShot  # type: ignore[assignment]
        mon1 = sct.monitors[1]
        image = sct.grab(mon1)
    assert isinstance(image, SimpleScreenShot)
    assert isinstance(image.raw, bytes)
    assert isinstance(image.monitor, dict)



"""This is part of the MSS Python's module.
Source: https://github.com/BoboTiG/python-mss.

2018-03-19.

Maximum screenshots in 1 second by computing BGRA raw values to RGB.


GNU/Linux
  pil_frombytes     139
  mss_rgb           119
  pil_frombytes_rgb 51
  numpy_flip        31
  numpy_slice       29

macOS
  pil_frombytes     209
  mss_rgb           174
  pil_frombytes_rgb 113
  numpy_flip        39
  numpy_slice       36

Windows
  pil_frombytes     81
  mss_rgb           66
  pil_frombytes_rgb 42
  numpy_flip        25
  numpy_slice       22
"""

import time

import numpy as np
from PIL import Image

import mss
from mss.screenshot import ScreenShot


def mss_rgb(im: ScreenShot) -> bytes:
    return im.rgb


def numpy_flip(im: ScreenShot) -> bytes:
    frame = np.array(im, dtype=np.uint8)
    return np.flip(frame[:, :, :3], 2).tobytes()


def numpy_slice(im: ScreenShot) -> bytes:
    return np.array(im, dtype=np.uint8)[..., [2, 1, 0]].tobytes()


def pil_frombytes_rgb(im: ScreenShot) -> bytes:
    return Image.frombytes("RGB", im.size, im.rgb).tobytes()


def pil_frombytes(im: ScreenShot) -> bytes:
    return Image.frombytes("RGB", im.size, im.bgra, "raw", "BGRX").tobytes()


def benchmark() -> None:
    with mss.mss() as sct:
        im = sct.grab(sct.monitors[0])
        for func in (
            pil_frombytes,
            mss_rgb,
            pil_frombytes_rgb,
            numpy_flip,
            numpy_slice,
        ):
            count = 0
            start = time.time()
            while (time.time() - start) <= 1:
                func(im)
                im._ScreenShot__rgb = None  # type: ignore[attr-defined]
                count += 1
            print(func.__name__.ljust(17), count)


benchmark()



"""Sharing Example

This example demonstrates how to easily share screenshots using PyShotter.
"""

from pyshotter import pyshotter, SharingFeature

def share_screenshot():
    """Demonstrate various sharing methods."""
    
    with pyshotter() as sct:
        # Take a screenshot
        screenshot = sct.grab(sct.monitors[0])
        
        # Initialize sharing feature
        sharer = SharingFeature()
        
        # Method 1: Copy to clipboard
        if sharer.copy_to_clipboard(screenshot):
            print("✅ Screenshot copied to clipboard!")
        else:
            print("❌ Clipboard copy failed (install pyperclip: pip install pyperclip)")
        
        # Method 2: Generate shareable link
        shareable_link = sharer.generate_shareable_link(screenshot)
        if shareable_link:
            print(f"🔗 Shareable link: {shareable_link}")
        else:
            print("❌ Failed to generate shareable link")
        
        # Method 3: Save with metadata
        metadata = {
            "title": "My Screenshot",
            "description": "Screenshot taken with PyShotter",
            "tags": ["python", "screenshot", "demo"],
            "author": "Abdoullah Ndao",
            "timestamp": "2024-12-19"
        }
        
        if sharer.save_with_metadata(screenshot, "screenshot_with_metadata.png", metadata):
            print("✅ Screenshot saved with metadata!")
        else:
            print("❌ Failed to save with metadata")

def share_annotated_screenshot():
    """Share an annotated screenshot."""
    
    from pyshotter import AnnotationFeature
    
    with pyshotter() as sct:
        screenshot = sct.grab(sct.monitors[0])
        
        # Add some annotations
        annotator = AnnotationFeature()
        annotated = annotator.add_text(screenshot, "Shared via PyShotter!", (100, 100))
        annotated = annotator.add_rectangle(annotated, (50, 50), (300, 200))
        
        # Share the annotated screenshot
        sharer = SharingFeature()
        
        # Copy to clipboard
        if sharer.copy_to_clipboard(annotated):
            print("✅ Annotated screenshot copied to clipboard!")
        
        # Generate link
        link = sharer.generate_shareable_link(annotated)
        print(f"🔗 Annotated screenshot link: {link}")

if __name__ == "__main__":
    print("📤 PyShotter Sharing Demo")
    print("=" * 40)
    
    share_screenshot()
    print("\n" + "=" * 40)
    share_annotated_screenshot() 


"""This is part of the MSS Python's module.
Source: https://github.com/BoboTiG/python-mss.
"""

import platform
import tarfile
from subprocess import STDOUT, check_call, check_output
from zipfile import ZipFile

import pytest

from pyshotter import __version__

if platform.system().lower() != "linux":
    pytestmark = pytest.mark.skip

pytest.importorskip("build")
pytest.importorskip("twine")

SDIST = "python -m build --sdist".split()
WHEEL = "python -m build --wheel".split()
CHECK = "twine check --strict".split()


def test_sdist() -> None:
    output = check_output(SDIST, stderr=STDOUT, text=True)
    file = f"pyshotter-{__version__}.tar.gz"
    assert f"Successfully built {file}" in output
    assert "warning" not in output.lower()

    check_call([*CHECK, f"dist/{file}"])

    with tarfile.open(f"dist/{file}", mode="r:gz") as fh:
        files = sorted(fh.getnames())

    assert files == [
        f"pyshotter-{__version__}/.gitignore",
        f"pyshotter-{__version__}/CHANGELOG.md",
        f"pyshotter-{__version__}/CHANGES.md",
        f"pyshotter-{__version__}/CONTRIBUTORS.md",
        f"pyshotter-{__version__}/LICENSE.txt",
        f"pyshotter-{__version__}/PKG-INFO",
        f"pyshotter-{__version__}/README.md",
        f"pyshotter-{__version__}/docs/source/api.rst",
        f"pyshotter-{__version__}/docs/source/conf.py",
        f"pyshotter-{__version__}/docs/source/developers.rst",
        f"pyshotter-{__version__}/docs/source/examples.rst",
        f"pyshotter-{__version__}/docs/source/examples/callback.py",
        f"pyshotter-{__version__}/docs/source/examples/custom_cls_image.py",
        f"pyshotter-{__version__}/docs/source/examples/fps.py",
        f"pyshotter-{__version__}/docs/source/examples/fps_multiprocessing.py",
        f"pyshotter-{__version__}/docs/source/examples/from_pil_tuple.py",
        f"pyshotter-{__version__}/docs/source/examples/linux_display_keyword.py",
        f"pyshotter-{__version__}/docs/source/examples/opencv_numpy.py",
        f"pyshotter-{__version__}/docs/source/examples/part_of_screen.py",
        f"pyshotter-{__version__}/docs/source/examples/part_of_screen_monitor_2.py",
        f"pyshotter-{__version__}/docs/source/examples/pil.py",
        f"pyshotter-{__version__}/docs/source/examples/pil_pixels.py",
        f"pyshotter-{__version__}/docs/source/index.rst",
        f"pyshotter-{__version__}/docs/source/installation.rst",
        f"pyshotter-{__version__}/docs/source/support.rst",
        f"pyshotter-{__version__}/docs/source/usage.rst",
        f"pyshotter-{__version__}/docs/source/where.rst",
        f"pyshotter-{__version__}/pyproject.toml",
        f"pyshotter-{__version__}/src/pyshotter/__init__.py",
        f"pyshotter-{__version__}/src/pyshotter/__main__.py",
        f"pyshotter-{__version__}/src/pyshotter/base.py",
        f"pyshotter-{__version__}/src/pyshotter/darwin.py",
        f"pyshotter-{__version__}/src/pyshotter/exception.py",
        f"pyshotter-{__version__}/src/pyshotter/factory.py",
        f"pyshotter-{__version__}/src/pyshotter/linux.py",
        f"pyshotter-{__version__}/src/pyshotter/models.py",
        f"pyshotter-{__version__}/src/pyshotter/py.typed",
        f"pyshotter-{__version__}/src/pyshotter/screenshot.py",
        f"pyshotter-{__version__}/src/pyshotter/tools.py",
        f"pyshotter-{__version__}/src/pyshotter/windows.py",
        f"pyshotter-{__version__}/src/tests/__init__.py",
        f"pyshotter-{__version__}/src/tests/bench_bgra2rgb.py",
        f"pyshotter-{__version__}/src/tests/bench_general.py",
        f"pyshotter-{__version__}/src/tests/conftest.py",
        f"pyshotter-{__version__}/src/tests/res/monitor-1024x768.raw.zip",
        f"pyshotter-{__version__}/src/tests/test_bgra_to_rgb.py",
        f"pyshotter-{__version__}/src/tests/test_cls_image.py",
        f"pyshotter-{__version__}/src/tests/test_find_monitors.py",
        f"pyshotter-{__version__}/src/tests/test_get_pixels.py",
        f"pyshotter-{__version__}/src/tests/test_gnu_linux.py",
        f"pyshotter-{__version__}/src/tests/test_implementation.py",
        f"pyshotter-{__version__}/src/tests/test_issue_220.py",
        f"pyshotter-{__version__}/src/tests/test_leaks.py",
        f"pyshotter-{__version__}/src/tests/test_macos.py",
        f"pyshotter-{__version__}/src/tests/test_save.py",
        f"pyshotter-{__version__}/src/tests/test_setup.py",
        f"pyshotter-{__version__}/src/tests/test_tools.py",
        f"pyshotter-{__version__}/src/tests/test_windows.py",
        f"pyshotter-{__version__}/src/tests/third_party/__init__.py",
        f"pyshotter-{__version__}/src/tests/third_party/test_numpy.py",
        f"pyshotter-{__version__}/src/tests/third_party/test_pil.py",
    ]


def test_wheel() -> None:
    output = check_output(WHEEL, stderr=STDOUT, text=True)
    file = f"pyshotter-{__version__}-py3-none-any.whl"
    assert f"Successfully built {file}" in output
    assert "warning" not in output.lower()

    check_call([*CHECK, f"dist/{file}"])

    with ZipFile(f"dist/{file}") as fh:
        files = sorted(fh.namelist())

    assert files == [
        f"pyshotter-{__version__}.dist-info/METADATA",
        f"pyshotter-{__version__}.dist-info/RECORD",
        f"pyshotter-{__version__}.dist-info/WHEEL",
        f"pyshotter-{__version__}.dist-info/entry_points.txt",
        f"pyshotter-{__version__}.dist-info/licenses/LICENSE.txt",
        "pyshotter/__init__.py",
        "pyshotter/__main__.py",
        "pyshotter/base.py",
        "pyshotter/darwin.py",
        "pyshotter/exception.py",
        "pyshotter/factory.py",
        "pyshotter/linux.py",
        "pyshotter/models.py",
        "pyshotter/py.typed",
        "pyshotter/screenshot.py",
        "pyshotter/tools.py",
        "pyshotter/windows.py",
    ]



"""Complete PyShotter Demo

This example demonstrates all of PyShotter's smart, annotated, and shareable features.
"""

import time
from pyshotter import (
    pyshotter, 
    SmartDetectionFeature,
    AnnotationFeature,
    SharingFeature,
    OCRFeature,
    RedactionFeature,
    PanoramaFeature,
    ChangeDetectionFeature,
    ScreenshotHistory
)

def complete_pyshotter_demo():
    """Demonstrate all PyShotter features: Smart, Annotated, and Shareable."""
    
    print("🚀 PyShotter: Smart, Annotated, and Shareable Screenshots")
    print("=" * 60)
    
    with pyshotter() as sct:
        # 1. SMART DETECTION
        print("\n🧠 STEP 1: Smart Detection")
        print("-" * 30)
        
        screenshot = sct.grab(sct.monitors[0])
        
        # Detect code regions and windows
        detector = SmartDetectionFeature()
        code_regions = detector.detect_code_regions(screenshot)
        windows = detector.detect_windows(screenshot)
        
        print(f"✅ Detected {len(code_regions)} code regions")
        print(f"✅ Detected {len(windows)} windows")
        
        # 2. OCR TEXT EXTRACTION
        print("\n📝 STEP 2: OCR Text Extraction")
        print("-" * 30)
        
        ocr = OCRFeature()
        text = ocr.extract_text(screenshot)
        print(f"✅ Extracted text: {text[:100]}...")
        
        # 3. ANNOTATION
        print("\n🎨 STEP 3: Rich Annotation")
        print("-" * 30)
        
        annotator = AnnotationFeature()
        annotated = screenshot
        
        # Add text annotation
        annotated = annotator.add_text(
            annotated, 
            "PyShotter Demo", 
            (50, 50),
            font_size=24,
            color=(255, 0, 0)
        )
        
        # Add rectangle around detected code regions
        for i, region in enumerate(code_regions[:3]):
            x, y, w, h = region['bbox']
            annotated = annotator.add_rectangle(
                annotated,
                (x, y),
                (x + w, y + h),
                color=(0, 255, 0),
                thickness=2
            )
            # Add label
            annotated = annotator.add_text(
                annotated,
                f"Code {i+1}",
                (x, y - 20),
                font_size=16,
                color=(0, 255, 0)
            )
        
        # Add arrow pointing to important area
        annotated = annotator.add_arrow(
            annotated,
            (100, 100),
            (300, 150),
            color=(0, 0, 255),
            thickness=3
        )
        
        # Add highlight
        annotated = annotator.add_highlight(
            annotated,
            (200, 200, 150, 100),
            color=(255, 255, 0),
            alpha=0.3
        )
        
        print("✅ Added text, shapes, arrows, and highlights")
        
        # 4. SHARING
        print("\n📤 STEP 4: Easy Sharing")
        print("-" * 30)
        
        sharer = SharingFeature()
        
        # Copy to clipboard
        if sharer.copy_to_clipboard(annotated):
            print("✅ Screenshot copied to clipboard!")
        
        # Generate shareable link
        link = sharer.generate_shareable_link(annotated)
        print(f"✅ Shareable link generated: {link[:50]}...")
        
        # Save with metadata
        metadata = {
            "title": "PyShotter Demo Screenshot",
            "description": "Smart, annotated, and shareable screenshot",
            "tags": ["demo", "pyshotter", "smart", "annotated", "shareable"],
            "author": "Abdoullah Ndao",
            "features_used": ["smart_detection", "annotation", "sharing"]
        }
        
        sharer.save_with_metadata(annotated, "demo_screenshot.png", metadata)
        print("✅ Screenshot saved with metadata")
        
        # 5. ADVANCED FEATURES
        print("\n🔧 STEP 5: Advanced Features")
        print("-" * 30)
        
        # Redaction
        redaction = RedactionFeature()
        clean_screenshot = redaction.redact_sensitive_data(screenshot)
        clean_screenshot.save("redacted_screenshot.png")
        print("✅ Sensitive data redacted")
        
        # Panorama (if multiple monitors)
        if len(sct.monitors) > 1:
            panorama = PanoramaFeature()
            screenshots = [sct.grab(monitor) for monitor in sct.monitors]
            panoramic = panorama.create_panorama(screenshots)
            panoramic.save("panorama.png")
            print("✅ Multi-monitor panorama created")
        
        # Change detection
        time.sleep(1)  # Wait a bit
        current_screenshot = sct.grab(sct.monitors[0])
        change_detector = ChangeDetectionFeature()
        changes = change_detector.detect_changes(current_screenshot, screenshot)
        changes.save("changes.png")
        print("✅ Change detection completed")
        
        # History and search
        history = ScreenshotHistory()
        screenshot_id = history.add_screenshot(
            annotated,
            metadata={
                "tags": ["demo", "complete"],
                "description": "Complete PyShotter demo screenshot"
            }
        )
        print(f"✅ Added to history with ID: {screenshot_id}")
        
        # Search history
        results = history.search_history("demo")
        print(f"✅ Found {len(results)} screenshots in history")
        
        # 6. FINAL RESULT
        print("\n🎉 STEP 6: Final Result")
        print("-" * 30)
        
        # Save the final annotated screenshot
        annotated.save("complete_demo_result.png")
        
        print("✅ Complete demo finished!")
        print("\n📁 Generated files:")
        print("  - complete_demo_result.png (annotated)")
        print("  - redacted_screenshot.png (clean)")
        print("  - changes.png (change detection)")
        if len(sct.monitors) > 1:
            print("  - panorama.png (multi-monitor)")
        print("  - demo_screenshot.png (with metadata)")
        print("  - History saved to ~/.pyshotter/history/")
        
        print("\n🚀 PyShotter makes screenshots smart, annotated, and shareable!")

if __name__ == "__main__":
    complete_pyshotter_demo() 


"""This is part of the MSS Python's module.
Source: https://github.com/BoboTiG/python-mss.

OpenCV/Numpy example.
"""

import time

import cv2
import numpy as np

import mss

with mss.mss() as sct:
    # Part of the screen to capture
    monitor = {"top": 40, "left": 0, "width": 800, "height": 640}

    while "Screen capturing":
        last_time = time.time()

        # Get raw pixels from the screen, save it to a Numpy array
        img = np.array(sct.grab(monitor))

        # Display the picture
        cv2.imshow("OpenCV/Numpy normal", img)

        # Display the picture in grayscale
        # cv2.imshow('OpenCV/Numpy grayscale',
        #            cv2.cvtColor(img, cv2.COLOR_BGRA2GRAY))

        print(f"fps: {1 / (time.time() - last_time)}")

        # Press "q" to quit
        if cv2.waitKey(25) & 0xFF == ord("q"):
            cv2.destroyAllWindows()
            break



"""This is part of the MSS Python's module.
Source: https://github.com/BoboTiG/python-mss.

PIL examples to play with pixels.
"""

from PIL import Image

import mss

with mss.mss() as sct:
    # Get a screenshot of the 1st monitor
    sct_img = sct.grab(sct.monitors[1])

    # Create an Image
    img = Image.new("RGB", sct_img.size)

    # Best solution: create a list(tuple(R, G, B), ...) for putdata()
    pixels = zip(sct_img.raw[2::4], sct_img.raw[1::4], sct_img.raw[::4])
    img.putdata(list(pixels))

    # But you can set individual pixels too (slower)
    """
    pixels = img.load()
    for x in range(sct_img.width):
        for y in range(sct_img.height):
            pixels[x, y] = sct_img.pixel(x, y)
    """

    # Show it!
    img.show()



"""This is part of the MSS Python's module.
Source: https://github.com/BoboTiG/python-mss.
"""

import platform
from collections.abc import Generator
from unittest.mock import Mock, patch

import pytest

import mss
import mss.linux
from mss.base import MSSBase
from mss.exception import ScreenShotError

pyvirtualdisplay = pytest.importorskip("pyvirtualdisplay")

PYPY = platform.python_implementation() == "PyPy"

WIDTH = 200
HEIGHT = 200
DEPTH = 24


@pytest.fixture
def display() -> Generator:
    with pyvirtualdisplay.Display(size=(WIDTH, HEIGHT), color_depth=DEPTH) as vdisplay:
        yield vdisplay.new_display_var


@pytest.mark.skipif(PYPY, reason="Failure on PyPy")
def test_factory_systems(monkeypatch: pytest.MonkeyPatch) -> None:
    """Here, we are testing all systems.

    Too hard to maintain the test for all platforms,
    so test only on GNU/Linux.
    """
    # GNU/Linux
    monkeypatch.setattr(platform, "system", lambda: "LINUX")
    with mss.mss() as sct:
        assert isinstance(sct, MSSBase)
    monkeypatch.undo()

    # macOS
    monkeypatch.setattr(platform, "system", lambda: "Darwin")
    # ValueError on macOS Big Sur
    with pytest.raises((ScreenShotError, ValueError)), mss.mss():
        pass
    monkeypatch.undo()

    # Windows
    monkeypatch.setattr(platform, "system", lambda: "wInDoWs")
    with pytest.raises(ImportError, match="cannot import name 'WINFUNCTYPE'"), mss.mss():
        pass


def test_arg_display(display: str, monkeypatch: pytest.MonkeyPatch) -> None:
    # Good value
    with mss.mss(display=display):
        pass

    # Bad `display` (missing ":" in front of the number)
    with pytest.raises(ScreenShotError), mss.mss(display="0"):
        pass

    # Invalid `display` that is not trivially distinguishable.
    with pytest.raises(ScreenShotError), mss.mss(display=":INVALID"):
        pass

    # No `DISPLAY` in envars
    monkeypatch.delenv("DISPLAY")
    with pytest.raises(ScreenShotError), mss.mss():
        pass


@pytest.mark.skipif(PYPY, reason="Failure on PyPy")
def test_bad_display_structure(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(mss.linux, "Display", lambda: None)
    with pytest.raises(TypeError), mss.mss():
        pass


@patch("mss.linux._X11", new=None)
def test_no_xlib_library() -> None:
    with pytest.raises(ScreenShotError), mss.mss():
        pass


@patch("mss.linux._XRANDR", new=None)
def test_no_xrandr_extension() -> None:
    with pytest.raises(ScreenShotError), mss.mss():
        pass


@patch("mss.linux.MSS._is_extension_enabled", new=Mock(return_value=False))
def test_xrandr_extension_exists_but_is_not_enabled(display: str) -> None:
    with pytest.raises(ScreenShotError), mss.mss(display=display):
        pass


def test_unsupported_depth() -> None:
    with (
        pyvirtualdisplay.Display(size=(WIDTH, HEIGHT), color_depth=8) as vdisplay,
        pytest.raises(ScreenShotError),
        mss.mss(display=vdisplay.new_display_var) as sct,
    ):
        sct.grab(sct.monitors[1])


def test_region_out_of_monitor_bounds(display: str) -> None:
    monitor = {"left": -30, "top": 0, "width": WIDTH, "height": HEIGHT}

    assert not mss.linux._ERROR

    with mss.mss(display=display) as sct:
        with pytest.raises(ScreenShotError) as exc:
            sct.grab(monitor)

        assert str(exc.value)

        details = exc.value.details
        assert details
        assert isinstance(details, dict)
        assert isinstance(details["error"], str)
        assert not mss.linux._ERROR

    assert not mss.linux._ERROR


def test__is_extension_enabled_unknown_name(display: str) -> None:
    with mss.mss(display=display) as sct:
        assert isinstance(sct, mss.linux.MSS)  # For Mypy
        assert not sct._is_extension_enabled("NOEXT")


def test_missing_fast_function_for_monitor_details_retrieval(display: str) -> None:
    with mss.mss(display=display) as sct:
        assert isinstance(sct, mss.linux.MSS)  # For Mypy
        assert hasattr(sct.xrandr, "XRRGetScreenResourcesCurrent")
        screenshot_with_fast_fn = sct.grab(sct.monitors[1])

    assert set(screenshot_with_fast_fn.rgb) == {0}

    with mss.mss(display=display) as sct:
        assert isinstance(sct, mss.linux.MSS)  # For Mypy
        assert hasattr(sct.xrandr, "XRRGetScreenResourcesCurrent")
        del sct.xrandr.XRRGetScreenResourcesCurrent
        screenshot_with_slow_fn = sct.grab(sct.monitors[1])

    assert set(screenshot_with_slow_fn.rgb) == {0}


def test_with_cursor(display: str) -> None:
    with mss.mss(display=display) as sct:
        assert not hasattr(sct, "xfixes")
        assert not sct.with_cursor
        screenshot_without_cursor = sct.grab(sct.monitors[1])

    # 1 color: black
    assert set(screenshot_without_cursor.rgb) == {0}

    with mss.mss(display=display, with_cursor=True) as sct:
        assert hasattr(sct, "xfixes")
        assert sct.with_cursor
        screenshot_with_cursor = sct.grab(sct.monitors[1])

    # 2 colors: black & white (default cursor is a white cross)
    assert set(screenshot_with_cursor.rgb) == {0, 255}


@patch("mss.linux._XFIXES", new=None)
def test_with_cursor_but_not_xfixes_extension_found(display: str) -> None:
    with mss.mss(display=display, with_cursor=True) as sct:
        assert not hasattr(sct, "xfixes")
        assert not sct.with_cursor


def test_with_cursor_failure(display: str) -> None:
    with mss.mss(display=display, with_cursor=True) as sct:
        assert isinstance(sct, mss.linux.MSS)  # For Mypy
        with (
            patch.object(sct.xfixes, "XFixesGetCursorImage", return_value=None),
            pytest.raises(ScreenShotError),
        ):
            sct.grab(sct.monitors[1])



"""This is part of the MSS Python's module.
Source: https://github.com/BoboTiG/python-mss.
"""

import pytest

import mss

tkinter = pytest.importorskip("tkinter")


@pytest.fixture
def root() -> tkinter.Tk:  # type: ignore[name-defined]
    try:
        master = tkinter.Tk()
    except RuntimeError:
        pytest.skip(reason="tk.h version (8.5) doesn't match libtk.a version (8.6)")

    try:
        yield master
    finally:
        master.destroy()


def take_screenshot() -> None:
    region = {"top": 370, "left": 1090, "width": 80, "height": 390}
    with mss.mss() as sct:
        sct.grab(region)


def create_top_level_win(master: tkinter.Tk) -> None:  # type: ignore[name-defined]
    top_level_win = tkinter.Toplevel(master)

    take_screenshot_btn = tkinter.Button(top_level_win, text="Take screenshot", command=take_screenshot)
    take_screenshot_btn.pack()

    take_screenshot_btn.invoke()
    master.update_idletasks()
    master.update()

    top_level_win.destroy()
    master.update_idletasks()
    master.update()


def test_regression(root: tkinter.Tk, capsys: pytest.CaptureFixture) -> None:  # type: ignore[name-defined]
    btn = tkinter.Button(root, text="Open TopLevel", command=lambda: create_top_level_win(root))
    btn.pack()

    # First screenshot: it works
    btn.invoke()

    # Second screenshot: it should work too
    btn.invoke()

    # Check there were no exceptions
    captured = capsys.readouterr()
    assert not captured.out
    assert not captured.err



"""This is part of the MSS Python's module.
Source: https://github.com/BoboTiG/python-mss.
"""

import os

from mss import mss


def test_get_monitors() -> None:
    with mss(display=os.getenv("DISPLAY")) as sct:
        assert sct.monitors


def test_keys_aio() -> None:
    with mss(display=os.getenv("DISPLAY")) as sct:
        all_monitors = sct.monitors[0]
    assert "top" in all_monitors
    assert "left" in all_monitors
    assert "height" in all_monitors
    assert "width" in all_monitors


def test_keys_monitor_1() -> None:
    with mss(display=os.getenv("DISPLAY")) as sct:
        mon1 = sct.monitors[1]
    assert "top" in mon1
    assert "left" in mon1
    assert "height" in mon1
    assert "width" in mon1


def test_dimensions() -> None:
    with mss(display=os.getenv("DISPLAY")) as sct:
        mon = sct.monitors[1]
    assert mon["width"] > 0
    assert mon["height"] > 0



# Lets prevent misses, and import the module to get the proper version.
# So that the version in only defined once across the whole code base:
#   src/pyshotter/__init__.py
import sys
from pathlib import Path

sys.path.insert(0, str(Path(__file__).parent.parent.parent / "src"))

import pyshotter

# -- General configuration ------------------------------------------------

extensions = [
    "sphinx_copybutton",
    "sphinx.ext.intersphinx",
    "sphinx_new_tab_link",
]
templates_path = ["_templates"]
source_suffix = {".rst": "restructuredtext"}
master_doc = "index"
new_tab_link_show_external_link_icon = True

# General information about the project.
project = "PyShotter"
copyright = f"{pyshotter.__date__}, {pyshotter.__author__} & contributors"  # noqa:A001
author = pyshotter.__author__
version = pyshotter.__version__

release = "latest"
language = "en"
todo_include_todos = True


# -- Options for HTML output ----------------------------------------------

html_theme = "shibuya"
html_theme_options = {
    "accent_color": "lime",
    "globaltoc_expand_depth": 1,
    "toctree_titles_only": False,
}
html_favicon = "../icon.png"
html_context = {
    "source_type": "github",
    "source_user": "utachicodes",
    "source_repo": "pyshotter",
    "source_docs_path": "/docs/source/",
    "source_version": "main",
}
htmlhelp_basename = "PyShotterdoc"


# -- Options for Epub output ----------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project
epub_author = author
epub_publisher = author
epub_copyright = copyright

# A list of files that should not be packed into the epub file.
epub_exclude_files = ["search.html"]


# ----------------------------------------------

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {"python": ("https://docs.python.org/3", None)}



"""This is part of the MSS Python's module.
Source: https://github.com/BoboTiG/python-mss.
"""

from __future__ import annotations

import os
import os.path
import platform
import sys
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING
from unittest.mock import Mock, patch

import pytest

import mss
import mss.tools
from mss.__main__ import main as entry_point
from mss.base import MSSBase
from mss.exception import ScreenShotError
from mss.screenshot import ScreenShot

if TYPE_CHECKING:  # pragma: nocover
    from mss.models import Monitor

try:
    from datetime import UTC
except ImportError:
    # Python < 3.11
    from datetime import timezone

    UTC = timezone.utc


class MSS0(MSSBase):
    """Nothing implemented."""


class MSS1(MSSBase):
    """Only `grab()` implemented."""

    def grab(self, monitor: Monitor) -> None:  # type: ignore[override]
        pass


class MSS2(MSSBase):
    """Only `monitor` implemented."""

    @property
    def monitors(self) -> list:
        return []


@pytest.mark.parametrize("cls", [MSS0, MSS1, MSS2])
def test_incomplete_class(cls: type[MSSBase]) -> None:
    with pytest.raises(TypeError):
        cls()


def test_bad_monitor() -> None:
    with mss.mss(display=os.getenv("DISPLAY")) as sct, pytest.raises(ScreenShotError):
        sct.shot(mon=222)


def test_repr() -> None:
    box = {"top": 0, "left": 0, "width": 10, "height": 10}
    expected_box = {"top": 0, "left": 0, "width": 10, "height": 10}
    with mss.mss(display=os.getenv("DISPLAY")) as sct:
        img = sct.grab(box)
    ref = ScreenShot(bytearray(b"42"), expected_box)
    assert repr(img) == repr(ref)


def test_factory(monkeypatch: pytest.MonkeyPatch) -> None:
    # Current system
    with mss.mss() as sct:
        assert isinstance(sct, MSSBase)

    # Unknown
    monkeypatch.setattr(platform, "system", lambda: "Chuck Norris")
    with pytest.raises(ScreenShotError) as exc:
        mss.mss()
    monkeypatch.undo()

    error = exc.value.args[0]
    assert error == "System 'chuck norris' not (yet?) implemented."


@patch.object(sys, "argv", new=[])  # Prevent side effects while testing
@pytest.mark.parametrize("with_cursor", [False, True])
def test_entry_point(with_cursor: bool, capsys: pytest.CaptureFixture) -> None:
    def main(*args: str, ret: int = 0) -> None:
        if with_cursor:
            args = (*args, "--with-cursor")
        assert entry_point(*args) == ret

    # No arguments
    main()
    captured = capsys.readouterr()
    for mon, line in enumerate(captured.out.splitlines(), 1):
        filename = Path(f"monitor-{mon}.png")
        assert line.endswith(filename.name)
        assert filename.is_file()
        filename.unlink()

    file = Path("monitor-1.png")
    for opt in ("-m", "--monitor"):
        main(opt, "1")
        captured = capsys.readouterr()
        assert captured.out.endswith(f"{file.name}\n")
        assert filename.is_file()
        filename.unlink()

    for opts in zip(["-m 1", "--monitor=1"], ["-q", "--quiet"]):
        main(*opts)
        captured = capsys.readouterr()
        assert not captured.out
        assert filename.is_file()
        filename.unlink()

    fmt = "sct-{mon}-{width}x{height}.png"
    for opt in ("-o", "--out"):
        main(opt, fmt)
        captured = capsys.readouterr()
        with mss.mss(display=os.getenv("DISPLAY")) as sct:
            for mon, (monitor, line) in enumerate(zip(sct.monitors[1:], captured.out.splitlines()), 1):
                filename = Path(fmt.format(mon=mon, **monitor))
                assert line.endswith(filename.name)
                assert filename.is_file()
                filename.unlink()

    fmt = "sct_{mon}-{date:%Y-%m-%d}.png"
    for opt in ("-o", "--out"):
        main("-m 1", opt, fmt)
        filename = Path(fmt.format(mon=1, date=datetime.now(tz=UTC)))
        captured = capsys.readouterr()
        assert captured.out.endswith(f"{filename}\n")
        assert filename.is_file()
        filename.unlink()

    coordinates = "2,12,40,67"
    filename = Path("sct-2x12_40x67.png")
    for opt in ("-c", "--coordinates"):
        main(opt, coordinates)
        captured = capsys.readouterr()
        assert captured.out.endswith(f"{filename}\n")
        assert filename.is_file()
        filename.unlink()

    coordinates = "2,12,40"
    for opt in ("-c", "--coordinates"):
        main(opt, coordinates, ret=2)
        captured = capsys.readouterr()
        assert captured.out == "Coordinates syntax: top, left, width, height\n"


@patch.object(sys, "argv", new=[])  # Prevent side effects while testing
@patch("mss.base.MSSBase.monitors", new=[])
@pytest.mark.parametrize("quiet", [False, True])
def test_entry_point_error(quiet: bool, capsys: pytest.CaptureFixture) -> None:
    def main(*args: str) -> int:
        if quiet:
            args = (*args, "--quiet")
        return entry_point(*args)

    if quiet:
        assert main() == 1
        captured = capsys.readouterr()
        assert not captured.out
        assert not captured.err
    else:
        with pytest.raises(ScreenShotError):
            main()


def test_entry_point_with_no_argument(capsys: pytest.CaptureFixture) -> None:
    # Make sure to fail if arguments are not handled
    with (
        patch("mss.factory.mss", new=Mock(side_effect=RuntimeError("Boom!"))),
        patch.object(sys, "argv", ["mss", "--help"]),
        pytest.raises(SystemExit) as exc,
    ):
        entry_point()
    assert exc.value.code == 0

    captured = capsys.readouterr()
    assert not captured.err
    assert "usage: mss" in captured.out


def test_grab_with_tuple() -> None:
    left = 100
    top = 100
    right = 500
    lower = 500
    width = right - left  # 400px width
    height = lower - top  # 400px height

    with mss.mss(display=os.getenv("DISPLAY")) as sct:
        # PIL like
        box = (left, top, right, lower)
        im = sct.grab(box)
        assert im.size == (width, height)

        # MSS like
        box2 = {"left": left, "top": top, "width": width, "height": height}
        im2 = sct.grab(box2)
        assert im.size == im2.size
        assert im.pos == im2.pos
        assert im.rgb == im2.rgb


def test_grab_with_tuple_percents() -> None:
    with mss.mss(display=os.getenv("DISPLAY")) as sct:
        monitor = sct.monitors[1]
        left = monitor["left"] + monitor["width"] * 5 // 100  # 5% from the left
        top = monitor["top"] + monitor["height"] * 5 // 100  # 5% from the top
        right = left + 500  # 500px
        lower = top + 500  # 500px
        width = right - left
        height = lower - top

        # PIL like
        box = (left, top, right, lower)
        im = sct.grab(box)
        assert im.size == (width, height)

        # MSS like
        box2 = {"left": left, "top": top, "width": width, "height": height}
        im2 = sct.grab(box2)
        assert im.size == im2.size
        assert im.pos == im2.pos
        assert im.rgb == im2.rgb


def test_thread_safety() -> None:
    """Regression test for issue #169."""
    import threading
    import time

    def record(check: dict) -> None:
        """Record for one second."""
        start_time = time.time()
        while time.time() - start_time < 1:
            with mss.mss() as sct:
                sct.grab(sct.monitors[1])

        check[threading.current_thread()] = True

    checkpoint: dict = {}
    t1 = threading.Thread(target=record, args=(checkpoint,))
    t2 = threading.Thread(target=record, args=(checkpoint,))

    t1.start()
    time.sleep(0.5)
    t2.start()

    t1.join()
    t2.join()

    assert len(checkpoint) == 2



"""This is part of the MSS Python's module.
Source: https://github.com/BoboTiG/python-mss.
"""

import ctypes.util
import platform
from unittest.mock import patch

import pytest

import mss
from mss.exception import ScreenShotError

if platform.system().lower() != "darwin":
    pytestmark = pytest.mark.skip

import mss.darwin


def test_repr() -> None:
    # CGPoint
    point = mss.darwin.CGPoint(2.0, 1.0)
    ref1 = mss.darwin.CGPoint()
    ref1.x = 2.0
    ref1.y = 1.0
    assert repr(point) == repr(ref1)

    # CGSize
    size = mss.darwin.CGSize(2.0, 1.0)
    ref2 = mss.darwin.CGSize()
    ref2.width = 2.0
    ref2.height = 1.0
    assert repr(size) == repr(ref2)

    # CGRect
    rect = mss.darwin.CGRect(point, size)
    ref3 = mss.darwin.CGRect()
    ref3.origin.x = 2.0
    ref3.origin.y = 1.0
    ref3.size.width = 2.0
    ref3.size.height = 1.0
    assert repr(rect) == repr(ref3)


def test_implementation(monkeypatch: pytest.MonkeyPatch) -> None:
    # No `CoreGraphics` library
    version = float(".".join(platform.mac_ver()[0].split(".")[:2]))

    if version < 10.16:
        monkeypatch.setattr(ctypes.util, "find_library", lambda _: None)
        with pytest.raises(ScreenShotError):
            mss.mss()
        monkeypatch.undo()

    with mss.mss() as sct:
        assert isinstance(sct, mss.darwin.MSS)  # For Mypy

        # Test monitor's rotation
        original = sct.monitors[1]
        monkeypatch.setattr(sct.core, "CGDisplayRotation", lambda _: -90.0)
        sct._monitors = []
        modified = sct.monitors[1]
        assert original["width"] == modified["height"]
        assert original["height"] == modified["width"]
        monkeypatch.undo()

        # Test bad data retrieval
        monkeypatch.setattr(sct.core, "CGWindowListCreateImage", lambda *_: None)
        with pytest.raises(ScreenShotError):
            sct.grab(sct.monitors[1])


def test_scaling_on() -> None:
    """Screnshots are taken at the nominal resolution by default, but scaling can be turned on manually."""
    # Grab a 1x1 screenshot
    region = {"top": 0, "left": 0, "width": 1, "height": 1}

    with mss.mss() as sct:
        # Nominal resolution, i.e.: scaling is off
        assert sct.grab(region).size[0] == 1

        # Retina resolution, i.e.: scaling is on
        with patch.object(mss.darwin, "IMAGE_OPTIONS", 0):
            assert sct.grab(region).size[0] in {1, 2}  # 1 on the CI, 2 for all other the world



"""This is part of the MSS Python's module.
Source: https://github.com/BoboTiG/python-mss.

Example using the multiprocessing module to speed-up screen capture.
    https://github.com/pythonlessons/TensorFlow-object-detection-tutorial
"""

from multiprocessing import Process, Queue

import mss
import mss.tools


def grab(queue: Queue) -> None:
    rect = {"top": 0, "left": 0, "width": 600, "height": 800}

    with mss.mss() as sct:
        for _ in range(1_000):
            queue.put(sct.grab(rect))

    # Tell the other worker to stop
    queue.put(None)


def save(queue: Queue) -> None:
    number = 0
    output = "screenshots/file_{}.png"
    to_png = mss.tools.to_png

    while "there are screenshots":
        img = queue.get()
        if img is None:
            break

        to_png(img.rgb, img.size, output=output.format(number))
        number += 1


if __name__ == "__main__":
    # The screenshots queue
    queue: Queue = Queue()

    # 2 processes: one for grabbing and one for saving PNG files
    Process(target=grab, args=(queue,)).start()
    Process(target=save, args=(queue,)).start()



"""This is part of the PyShotter Python's module.
Source: https://github.com/utachicodes/pyshotter.
"""

import os

import numpy
import pytest

from pyshotter import pyshotter


def test_numpy():
    """Test that numpy can handle our images."""
    with pyshotter(display=os.getenv("DISPLAY")) as sct:
        sct_img = sct.grab(sct.monitors[0])
        arr = numpy.array(sct_img)
        assert arr.shape == (sct_img.height, sct_img.width, 3)



"""This is part of the MSS Python's module.
Source: https://github.com/BoboTiG/python-mss.

Simple naive benchmark to compare with:
    https://pythonprogramming.net/game-frames-open-cv-python-plays-gta-v/
"""

import time

import cv2
import numpy as np
from PIL import ImageGrab

import mss


def screen_record() -> int:
    # 800x600 windowed mode
    mon = (0, 40, 800, 640)

    title = "[PIL.ImageGrab] FPS benchmark"
    fps = 0
    last_time = time.time()

    while time.time() - last_time < 1:
        img = np.asarray(ImageGrab.grab(bbox=mon))
        fps += 1

        cv2.imshow(title, cv2.cvtColor(img, cv2.COLOR_BGR2RGB))
        if cv2.waitKey(25) & 0xFF == ord("q"):
            cv2.destroyAllWindows()
            break

    return fps


def screen_record_efficient() -> int:
    # 800x600 windowed mode
    mon = {"top": 40, "left": 0, "width": 800, "height": 640}

    title = "[MSS] FPS benchmark"
    fps = 0
    sct = mss.mss()
    last_time = time.time()

    while time.time() - last_time < 1:
        img = np.asarray(sct.grab(mon))
        fps += 1

        cv2.imshow(title, img)
        if cv2.waitKey(25) & 0xFF == ord("q"):
            cv2.destroyAllWindows()
            break

    return fps


print("PIL:", screen_record())
print("MSS:", screen_record_efficient())



"""This is part of the MSS Python's module.
Source: https://github.com/BoboTiG/python-mss.

2018-03-19.

Original means MSS 3.1.2.
Patched  means MSS 3.2.0.


GNU/Linux    Original Patched   Gain %
  grab           2618    2738    +4.58
  access_rgb     1083    1128    +4.15
  output          324     322   ------
  save            320     319   ------

macOS
  grab            524     526   ------
  access_rgb      396     406    +2.52
  output          194     195   ------
  save            193     194   ------

Windows
  grab           1280    2498   +95.16
  access_rgb      574     712   +24.04
  output          139     188   +35.25
"""

from __future__ import annotations

from time import time
from typing import TYPE_CHECKING

import mss
import mss.tools

if TYPE_CHECKING:  # pragma: nocover
    from collections.abc import Callable

    from mss.base import MSSBase
    from mss.screenshot import ScreenShot


def grab(sct: MSSBase) -> ScreenShot:
    monitor = {"top": 144, "left": 80, "width": 1397, "height": 782}
    return sct.grab(monitor)


def access_rgb(sct: MSSBase) -> bytes:
    im = grab(sct)
    return im.rgb


def output(sct: MSSBase, filename: str | None = None) -> None:
    rgb = access_rgb(sct)
    mss.tools.to_png(rgb, (1397, 782), output=filename)


def save(sct: MSSBase) -> None:
    output(sct, filename="screenshot.png")


def benchmark(func: Callable) -> None:
    count = 0
    start = time()

    with mss.mss() as sct:
        while (time() - start) % 60 < 10:
            count += 1
            func(sct)

    print(func.__name__, count)


benchmark(grab)
benchmark(access_rgb)
benchmark(output)
benchmark(save)



"""This is part of the MSS Python's module.
Source: https://github.com/BoboTiG/python-mss.
"""

import pytest

from mss.base import ScreenShot


def test_bad_length() -> None:
    data = bytearray(b"789c626001000000ffff030000060005")
    image = ScreenShot.from_size(data, 1024, 768)
    with pytest.raises(ValueError, match="attempt to assign"):
        _ = image.rgb


def test_good_types(raw: bytes) -> None:
    image = ScreenShot.from_size(bytearray(raw), 1024, 768)
    assert isinstance(image.raw, bytearray)
    assert isinstance(image.rgb, bytes)



"""This is part of the MSS Python's module.
Source: https://github.com/BoboTiG/python-mss.

Use PIL bbox style and percent values.
"""

import mss
import mss.tools

with mss.mss() as sct:
    # Use the 1st monitor
    monitor = sct.monitors[1]

    # Capture a bbox using percent values
    left = monitor["left"] + monitor["width"] * 5 // 100  # 5% from the left
    top = monitor["top"] + monitor["height"] * 5 // 100  # 5% from the top
    right = left + 400  # 400px width
    lower = top + 400  # 400px height
    bbox = (left, top, right, lower)

    # Grab the picture
    # Using PIL would be something like:
    # im = ImageGrab(bbox=bbox)
    im = sct.grab(bbox)

    # Save it!
    mss.tools.to_png(im.rgb, im.size, output="screenshot.png")



"""PyShotter: Smart, annotated, and shareable screenshots for Python.

Features:
- Smart detection of code regions and windows
- Rich annotation tools (text, shapes, highlights)
- Easy sharing (clipboard, links, cloud)
- OCR Text Extraction
- Sensitive Data Redaction
- Multi-Monitor Panorama
- Change Detection
- Customizable Hotkeys
- Screenshot History & Search

See: https://pyshotter.readthedocs.io
"""

from .exception import ScreenShotError
from .factory import pyshotter
from .features import (
    AnnotationFeature,
    SharingFeature,
    SmartDetectionFeature,
    OCRFeature,
    RedactionFeature,
    PanoramaFeature,
    ChangeDetectionFeature,
    HotkeyManager,
    ScreenshotHistory,
)

__version__ = "1.0.0"
__author__ = "Abdoullah Ndao"
__email__ = "abdoullahaljersi@gmail.com"
__date__ = "2024"
__copyright__ = "Copyright (c) 2024-2025, Abdoullah Ndao"

__all__ = (
    "ScreenShotError",
    "pyshotter",
    "AnnotationFeature",
    "SharingFeature", 
    "SmartDetectionFeature",
    "OCRFeature",
    "RedactionFeature", 
    "PanoramaFeature",
    "ChangeDetectionFeature",
    "HotkeyManager",
    "ScreenshotHistory",
)



"""This is part of the MSS Python's module.
Source: https://github.com/BoboTiG/python-mss.
"""

from collections.abc import Generator
from hashlib import sha256
from pathlib import Path
from zipfile import ZipFile

import pytest


@pytest.fixture(autouse=True)
def _no_warnings(recwarn: pytest.WarningsRecorder) -> Generator:
    """Fail on warning."""
    yield

    warnings = [f"{warning.filename}:{warning.lineno} {warning.message}" for warning in recwarn]
    for warning in warnings:
        print(warning)
    assert not warnings


def purge_files() -> None:
    """Remove all generated files from previous runs."""
    for file in Path().glob("*.png"):
        print(f"Deleting {file} ...")
        file.unlink()

    for file in Path().glob("*.png.old"):
        print(f"Deleting {file} ...")
        file.unlink()


@pytest.fixture(scope="module", autouse=True)
def _before_tests() -> None:
    purge_files()


@pytest.fixture(scope="session")
def raw() -> bytes:
    file = Path(__file__).parent / "res" / "monitor-1024x768.raw.zip"
    with ZipFile(file) as fh:
        data = fh.read(file.with_suffix("").name)

    assert sha256(data).hexdigest() == "d86ed4366d5a882cfe1345de82c87b81aef9f9bf085f4c42acb6f63f3967eccd"
    return data



"""This is part of the MSS Python's module.
Source: https://github.com/BoboTiG/python-mss.
"""

from __future__ import annotations

import threading

import pytest

import pyshotter
from pyshotter.exception import ScreenShotError

try:
    import pyshotter.windows
except ImportError:
    pytestmark = pytest.mark.skip


def test_implementation(monkeypatch: pytest.MonkeyPatch) -> None:
    # Test bad data retrieval
    with pyshotter.mss() as sct:
        assert isinstance(sct, pyshotter.windows.MSS)  # For Mypy

        monkeypatch.setattr(sct.gdi32, "GetDIBits", lambda *_: 0)
        with pytest.raises(ScreenShotError):
            sct.shot()


def test_region_caching() -> None:
    """The region to grab is cached, ensure this is well-done."""
    with pyshotter.mss() as sct:
        assert isinstance(sct, pyshotter.windows.MSS)  # For Mypy

        # Grab the area 1
        region1 = {"top": 0, "left": 0, "width": 200, "height": 200}
        sct.grab(region1)
        bmp1 = id(sct._handles.bmp)

        # Grab the area 2, the cached BMP is used
        # Same sizes but different positions
        region2 = {"top": 200, "left": 200, "width": 200, "height": 200}
        sct.grab(region2)
        bmp2 = id(sct._handles.bmp)
        assert bmp1 == bmp2

        # Grab the area 2 again, the cached BMP is used
        sct.grab(region2)
        assert bmp2 == id(sct._handles.bmp)


def test_region_not_caching() -> None:
    """The region to grab is not bad cached previous grab."""
    grab1 = pyshotter.mss()
    grab2 = pyshotter.mss()

    assert isinstance(grab1, pyshotter.windows.MSS)  # For Mypy
    assert isinstance(grab2, pyshotter.windows.MSS)  # For Mypy

    region1 = {"top": 0, "left": 0, "width": 100, "height": 100}
    region2 = {"top": 0, "left": 0, "width": 50, "height": 1}
    grab1.grab(region1)
    bmp1 = id(grab1._handles.bmp)
    grab2.grab(region2)
    bmp2 = id(grab2._handles.bmp)
    assert bmp1 != bmp2

    # Grab the area 1, is not bad cached BMP previous grab the area 2
    grab1.grab(region1)
    bmp1 = id(grab1._handles.bmp)
    assert bmp1 != bmp2


def run_child_thread(loops: int) -> None:
    for _ in range(loops):
        with pyshotter.mss() as sct:  # New sct for every loop
            sct.grab(sct.monitors[1])


def test_thread_safety() -> None:
    """Thread safety test for issue #150.

    The following code will throw a ScreenShotError exception if thread-safety is not guaranteed.
    """
    # Let thread 1 finished ahead of thread 2
    thread1 = threading.Thread(target=run_child_thread, args=(30,))
    thread2 = threading.Thread(target=run_child_thread, args=(50,))
    thread1.start()
    thread2.start()
    thread1.join()
    thread2.join()


def run_child_thread_bbox(loops: int, bbox: tuple[int, int, int, int]) -> None:
    with pyshotter.mss() as sct:  # One sct for all loops
        for _ in range(loops):
            sct.grab(bbox)


def test_thread_safety_regions() -> None:
    """Thread safety test for different regions.

    The following code will throw a ScreenShotError exception if thread-safety is not guaranteed.
    """
    thread1 = threading.Thread(target=run_child_thread_bbox, args=(100, (0, 0, 100, 100)))
    thread2 = threading.Thread(target=run_child_thread_bbox, args=(100, (0, 0, 50, 1)))
    thread1.start()
    thread2.start()
    thread1.join()
    thread2.join()
